/*****************************************************************************
 * vaapi.c: VAAPI helpers for the ffmpeg decoder
 *****************************************************************************
 * Copyright (C) 2009-2011 Laurent Aimar
 * $Id: 82762a954d4e544d1c4735ef63ed89ffaa13272b $
 *
 * Authors: Laurent Aimar <fenrir_AT_ videolan _DOT_ org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_fourcc.h>
#include <vlc_picture.h>
#include <vlc_arrays.h>
#include <assert.h>

#include "avcodec.h"

#ifdef HAVE_LIBAVCODEC_AVCODEC_H
#   include <libavcodec/avcodec.h>
#   include <libavcodec/vaapi.h>
#else
#   include <avcodec.h>
#   include <vaapi.h>
#endif

#include <vlc_xlib.h>
#include <va/va.h>
#include <va/va_x11.h>

#include "va.h"
#include "vaapi.h"
#include "copy.h"

struct vlc_va_sys_t
{
    /* */
    vlc_va_conn_t *conn;
    VAConfigID    i_config_id;

    struct vaapi_context hw_ctx;

    /* */
    int          i_surface_count;
    int          i_surface_width;
    int          i_surface_height;
    vlc_fourcc_t i_surface_chroma;

    VAImage      image;
    copy_cache_t image_cache;
};

static void Close( vlc_va_t *p_va );

static unsigned int vaFallbackFourCC[] =
{
    VA_FOURCC( 'Y', 'V', '1', '2' ),
    VA_FOURCC( 'I', '4', '2', '0' ),
    VA_FOURCC( 'N', 'V', '1', '2' )
};

/* */
static bool HasProfile( VAProfile profile, VAProfile *list, const int list_size )
{
    for (int i = 0; i < list_size; i++)
    {
        if (list[i] == profile)
            return true;
    }
    return false;
}

static VAProfile *GetHwProfiles( vlc_va_conn_t *p_conn, int *count )
{
    int num = vaMaxNumProfiles( p_conn->p_display );
    VAProfile *p_list = calloc( num, sizeof( VAProfile ) );
    if ( !p_list )
        return NULL;

    VAStatus status = vaQueryConfigProfiles( p_conn->p_display, p_list, &num );
    if ( status != VA_STATUS_SUCCESS )
    {
        free( p_list );
        return NULL;
    }
    *count = num;
    return p_list;
}

static VAProfile GetSupportedProfile( vlc_va_conn_t *p_conn, const int i_codec_id )
{
    static const int mpeg2_profiles[] =
        { VAProfileMPEG2Main, VAProfileMPEG2Simple, -1 };
    static const int mpeg4_profiles[] =
        { VAProfileMPEG4Main, VAProfileMPEG4AdvancedSimple, VAProfileMPEG4Simple, -1 };
    static const int h264_profiles[] =
        { VAProfileH264High, VAProfileH264Main, VAProfileH264Baseline, -1 };
    static const int wmv3_profiles[] =
        { VAProfileVC1Main, VAProfileVC1Simple, -1 };
    static const int vc1_profiles[] =
        { VAProfileVC1Advanced, -1 };

    const int *profiles = NULL;
    switch( i_codec_id )
    {
    case CODEC_ID_MPEG1VIDEO:
    case CODEC_ID_MPEG2VIDEO: profiles = mpeg2_profiles; break;
    case CODEC_ID_MPEG4:      profiles = mpeg4_profiles; break;
    case CODEC_ID_WMV3:       profiles = wmv3_profiles; break;
    case CODEC_ID_VC1:        profiles = vc1_profiles; break;
    case CODEC_ID_H264:       profiles = h264_profiles; break;
    default:
        return -1;
    }

    /* Get supported profiles from HW */
    int count = 0;
    VAProfile *p_list = GetHwProfiles( p_conn, &count );
    if( !p_list )
        return -1;

    /* Select best match of highest possible quality */
    int i_profile = -1;
    bool b_supported = false;
    for (int i = 0; profiles[i] > -1; i++)
    {
        b_supported = HasProfile( profiles[i], p_list, count );
        if (b_supported)
        {
            i_profile = profiles[i];
            break;
        }
    }
    free(p_list);

    if (!b_supported)
        return -1;

    return i_profile;
}

static int CalculateSurfaceCount( const int i_codec_id )
{
    int i_count = -1;

    switch( i_codec_id )
    {
    case CODEC_ID_MPEG1VIDEO:
    case CODEC_ID_MPEG2VIDEO: i_count = 2+1; break;
    case CODEC_ID_MPEG4:      i_count = 2+1; break;
    case CODEC_ID_WMV3:       i_count = 2+1; break;
    case CODEC_ID_VC1:        i_count = 2+1; break;
    case CODEC_ID_H264:       i_count = 20+1; break;
    default:
        return -1;
    }

    return i_count * 2;
}

/* */
static int Open( vlc_va_t *p_external, int i_codec_id )
{
    int i_profile;

    vlc_va_sys_t *p_va = calloc(1, sizeof(*p_va));
    if (unlikeyly(p_va == NULL))
       return VLC_ENOMEM;

    /* */
    p_va->i_config_id  = VA_INVALID_ID;
    p_va->image.image_id = VA_INVALID_ID;

    p_va->i_surface_count = CalculateSurfaceCount(i_codec_id);
    if (p_va->i_surface_count < 0 )
        return VLC_EGENERIC;

    /* Create a VA display */
    p_va->conn = vlc_va_Initialize(NULL);
    if (!p_va->conn)
        goto error;

    p_va->conn->lock();

    /* Check if the selected profile is supported */
    i_profile = GetSupportedProfile( p_va->conn, i_codec_id );
    if (i_profile < 0)
        goto unlock;

    /* Create a VA configuration */
    VAConfigAttrib attrib;
    memset( &attrib, 0, sizeof(attrib) );
    attrib.type = VAConfigAttribRTFormat;
    if( vaGetConfigAttributes( p_va->conn->p_display,
                               (VAProfile) i_profile, VAEntrypointVLD, &attrib, 1 ) )
        goto unlock;

    /* Not sure what to do if not, I don't have a way to test */
    if( (attrib.value & VA_RT_FORMAT_YUV420) == 0 )
        goto unlock;

    if( vaCreateConfig( p_va->conn->p_display, (VAProfile) i_profile,
                        VAEntrypointVLD, &attrib, 1, &p_va->i_config_id ) )
    {
        p_va->i_config_id = VA_INVALID_ID;
        goto unlock;
    }

    if( asprintf( &p_va->va.description, "VA API version %d.%d",
                  p_va->conn->i_version_major, p_va->conn->i_version_minor ) < 0 )
        p_va->va.description = NULL;

    p_va->conn->unlock();

    p_external->sys = p_va;
    return VLC_SUCCESS;

unlock:
    p_va->conn->unlock();
error:
    Close(p_external);
    free(p_va);
    return VLC_EGENERIC;
}

static void DestroyDecodingContext( vlc_va_sys_t *p_va )
{
    p_va->conn->lock();
    if( p_va->image.image_id != VA_INVALID_ID )
    {
        CopyCleanCache( &p_va->image_cache );
        vaDestroyImage( p_va->conn->p_display, p_va->image.image_id );
    }
    else if( p_va->conn->b_supports_derive )
    {
        CopyCleanCache( &p_va->image_cache );
    }

    if( p_va->conn->i_context_id != VA_INVALID_ID )
        vaDestroyContext( p_va->conn->p_display, p_va->conn->i_context_id );

    vlc_va_conn_DestroySurfaces( p_va->conn );

    /* */
    p_va->conn->i_context_id = VA_INVALID_ID;
    p_va->image.image_id = VA_INVALID_ID;
    p_va->i_surface_width = 0;
    p_va->i_surface_height = 0;
    p_va->conn->unlock();
}

static int CreateDecodingContext( vlc_va_sys_t *p_va, void **pp_hw_ctx, vlc_fourcc_t *pi_chroma,
                                  int i_width, int i_height )
{
    assert( i_width > 0 && i_height > 0 );

    p_va->conn->lock();

    /* Create surfaces */
    bool ok = vlc_va_conn_CreateSurface( p_va->conn, i_width, i_height,
                                         VA_RT_FORMAT_YUV420, p_va->i_surface_count );
    if( !ok )
    {
        p_va->conn->unlock();
        return VLC_EGENERIC;
    }

    /* */
    p_va->image.image_id = VA_INVALID_ID;
    assert(p_va->conn->i_context_id == VA_INVALID_ID);

    /* Create a context */
    const unsigned int count = vlc_va_conn_SurfacesCount(p_va->conn);
    VASurfaceID *p_ids = calloc(count, sizeof(VASurfaceID));
    if( !p_ids )
        goto error;
    for( unsigned int i = 0; i < count; i++ )
    {
        p_ids[i] = p_va->conn->pool.p_surfaces[i].i_id;
    }
    if( vaCreateContext( p_va->conn->p_display, p_va->i_config_id,
                         i_width, i_height, VA_PROGRESSIVE,
                         p_ids, count,
                         &p_va->conn->i_context_id ) != VA_STATUS_SUCCESS )
    {
        free(p_ids);
        p_va->conn->i_context_id = VA_INVALID_ID;
        goto error;
    }
    free(p_ids);

    if( p_va->conn->b_supports_derive )
    {
        video_format_t fmt;
        if( vlc_va_get_video_format( &p_va->conn->derive_format, &fmt ) != VLC_SUCCESS )
            goto error;

        *pi_chroma = fmt.i_chroma;
    }
    else
    {
        /* Find and create a supported image chroma */
        bool b_supported = false;
        for( unsigned int i = 0; i < ARRAY_SIZE(vaFallbackFourCC); i++ )
        {
            video_format_t fmt;
            VAImageFormat  format;
            if( vlc_va_get_image_format(p_va->conn, vaFallbackFourCC[i],
                                        &format, &fmt) != VLC_SUCCESS )
                continue;

            if( vaCreateImage( p_va->conn->p_display, &format,
                               i_width, i_height, &p_va->image ) )
            {
                p_va->image.image_id = VA_INVALID_ID;
                continue;
            }
            /* Validate that vaGetImage works with this format */
            if( vaGetImage( p_va->conn->p_display,
                            p_va->conn->pool.p_surfaces[0].i_id,
                            0, 0, i_width, i_height,
                            p_va->image.image_id ) )
            {
                vaDestroyImage( p_va->conn->p_display, p_va->image.image_id );
                p_va->image.image_id = VA_INVALID_ID;
                continue;
            }

            *pi_chroma = fmt.i_chroma;
            b_supported = true;
            break;
        }
        if( !b_supported )
            goto error;
    }

    if( unlikely(CopyInitCache( &p_va->image_cache, i_width )) )
        goto error;

    /* Setup the ffmpeg hardware context */
    *pp_hw_ctx = &p_va->hw_ctx;

    memset( &p_va->hw_ctx, 0, sizeof(p_va->hw_ctx) );
    p_va->hw_ctx.display    = p_va->conn->p_display;
    p_va->hw_ctx.config_id  = p_va->i_config_id;
    p_va->hw_ctx.context_id = p_va->conn->i_context_id;

    /* */
    p_va->i_surface_chroma = *pi_chroma;
    p_va->i_surface_width = i_width;
    p_va->i_surface_height = i_height;

    p_va->conn->unlock();
    return VLC_SUCCESS;

error:
    p_va->conn->unlock();
    DestroyDecodingContext( p_va );
    return VLC_EGENERIC;
}

static int Setup( vlc_va_t *p_external, void **pp_hw_ctx, vlc_fourcc_t *pi_chroma,
                  int i_width, int i_height )
{
    vlc_va_sys_t *p_va = p_external->sys;

    if( p_va->i_surface_width == i_width &&
        p_va->i_surface_height == i_height )
    {
        *pp_hw_ctx = &p_va->hw_ctx;
        *pi_chroma = p_va->i_surface_chroma;
        return VLC_SUCCESS;
    }

    *pp_hw_ctx = NULL;
    *pi_chroma = 0;
    if( p_va->i_surface_width || p_va->i_surface_height )
        DestroyDecodingContext( p_va );

    if( i_width > 0 && i_height > 0 )
        return CreateDecodingContext( p_va, pp_hw_ctx, pi_chroma, i_width, i_height );

    return VLC_EGENERIC;
}
static int Extract( vlc_va_t *p_external, picture_t *p_picture, AVFrame *p_ff )
{
    vlc_va_sys_t *p_va = p_external->sys;

    if( !p_va->image_cache.buffer )
        return VLC_EGENERIC;

    p_va->conn->lock();

    VASurfaceID i_surface_id = (VASurfaceID)(uintptr_t)p_ff->data[3];

#if VA_CHECK_VERSION(0,31,0)
    if( vaSyncSurface( p_va->conn->p_display, i_surface_id ) )
#else
    if( vaSyncSurface( p_va->conn->p_display, p_va->conn->i_context_id, i_surface_id ) )
#endif
        goto error;

    if( p_va->conn->b_supports_derive )
    {
        if( vaDeriveImage( p_va->conn->p_display, i_surface_id,
                           &(p_va->image) ) != VA_STATUS_SUCCESS )
            goto error;
    }
    else
    {
        if( vaGetImage( p_va->conn->p_display, i_surface_id,
                        0, 0, p_va->i_surface_width, p_va->i_surface_height,
                        p_va->image.image_id ) )
            goto error;
    }

    void *p_base;
    if( vaMapBuffer( p_va->conn->p_display, p_va->image.buf, &p_base ) )
        goto error;

    const uint32_t i_fourcc = p_va->image.format.fourcc;
    if( i_fourcc == VA_FOURCC('Y','V','1','2') ||
        i_fourcc == VA_FOURCC('I','4','2','0') )
    {
        bool b_swap_uv = i_fourcc == VA_FOURCC('I','4','2','0');
        uint8_t *pp_plane[3];
        size_t  pi_pitch[3];

        for( int i = 0; i < 3; i++ )
        {
            const int i_src_plane = (b_swap_uv && i != 0) ?  (3 - i) : i;
            pp_plane[i] = (uint8_t*)p_base + p_va->image.offsets[i_src_plane];
            pi_pitch[i] = p_va->image.pitches[i_src_plane];
        }
        CopyFromYv12( p_picture, pp_plane, pi_pitch,
                      p_va->i_surface_width,
                      p_va->i_surface_height,
                      &p_va->image_cache );
    }
    else
    {
        assert( i_fourcc == VA_FOURCC('N','V','1','2') );
        uint8_t *pp_plane[2];
        size_t  pi_pitch[2];

        for( int i = 0; i < 2; i++ )
        {
            pp_plane[i] = (uint8_t*)p_base + p_va->image.offsets[i];
            pi_pitch[i] = p_va->image.pitches[i];
        }
        CopyFromNv12( p_picture, pp_plane, pi_pitch,
                      p_va->i_surface_width,
                      p_va->i_surface_height,
                      &p_va->image_cache );
    }

    if( vaUnmapBuffer( p_va->conn->p_display, p_va->image.buf ) )
        goto error;

    if( p_va->conn->b_supports_derive )
    {
        vaDestroyImage( p_va->conn->p_display, p_va->image.image_id );
        p_va->image.image_id = VA_INVALID_ID;
    }

    p_va->conn->unlock();
    return VLC_SUCCESS;

error:
    p_va->conn->unlock();
    return VLC_EGENERIC;
}

static void Put(vlc_va_t *p_external, AVFrame *p_ff, picture_t *p_picture )
{
    vlc_va_sys_t *p_va = p_external->sys;

    p_va->conn->lock();

    if (!p_picture->p_sys)
        abort();

    picture_sys_t *p_surface = p_picture->p_sys;
    assert(p_surface->i_refcount == 2);

    /* */
    for( int i = 0; i < 4; i++ )
    {
        p_ff->data[i] = NULL;
        p_ff->linesize[i] = 0;

        if( i == 0 || i == 3 )
            p_ff->data[i] = (void*)(uintptr_t)p_surface->i_id;/* Yummie */
    }

    p_va->conn->unlock();
}

static int DisplayPicture( vlc_va_t *p_external, picture_t *p_picture )
{
    vlc_va_sys_t *p_va = p_external->sys;

    p_va->conn->lock();

    if (!p_picture->p_sys)
        abort();

    VASurfaceID i_surface_id = (VASurfaceID)p_picture->p_sys->i_id;

#if VA_CHECK_VERSION(0,31,0)
    if( vaSyncSurface( p_va->conn->p_display, i_surface_id ) )
#else
    if( vaSyncSurface( p_va->conn->p_display, p_va->conn->i_context_id, i_surface_id ) )
#endif
        goto error;

    p_va->conn->unlock();
    return VLC_SUCCESS;

error:
    assert(0);
    p_va->conn->unlock();
    return VLC_EGENERIC;
}

static int Get( vlc_va_t *p_external, AVFrame *p_ff )
{
    vlc_va_sys_t *p_va = p_external->sys;

    p_va->conn->lock();

    vlc_va_surface_t *p_surface = vlc_va_conn_GetSurface(p_va->conn);
    if (!p_surface)
    {
        msg_Warn(p_external, "No free surface found.");
        p_va->conn->unlock();
        return VLC_EGENERIC;
    }

    /* */
    vlc_mutex_lock(&p_surface->lock);
    for( int i = 0; i < 4; i++ )
    {
        p_ff->data[i] = NULL;
        p_ff->linesize[i] = 0;

        if( i == 0 || i == 3 )
            p_ff->data[i] = (void*)(uintptr_t)p_surface->i_id;/* Yummie */
    }
    vlc_mutex_unlock(&p_surface->lock);
    p_va->conn->unlock();

    return VLC_SUCCESS;
}

static void Release( vlc_va_t *p_external, AVFrame *p_ff )
{
    vlc_va_sys_t *p_va = p_external->sys;

    p_va->conn->lock();
    VASurfaceID i_surface_id = (VASurfaceID)(uintptr_t)p_ff->data[3];
    vlc_va_conn_ReleaseSurfaceById(p_va->conn, i_surface_id);
    p_va->conn->unlock();
}

static void Close( vlc_va_sys_t *p_va )
{
    if( p_va->i_surface_width || p_va->i_surface_height )
        DestroyDecodingContext( p_va );

    if (p_va->conn)
    {
        p_va->conn->lock();
        if( p_va->i_config_id != VA_INVALID_ID )
            vaDestroyConfig( p_va->conn->p_display, p_va->i_config_id );
        p_va->conn->unlock();

        vlc_va_Terminate( p_va->conn );
    }
}

static void Delete( vlc_va_t *p_external )
{
    vlc_va_sys_t *p_va = p_external->sys;
    Close( p_va );
    free( p_external->description );
    free( p_va );
}

static bool QuerySurfaceReady( vlc_va_t *p_external, picture_t *pic )
{
    vlc_va_sys_t *p_va = p_external->p_sys;

    assert(pic);
    assert(pic->p_sys);
    assert(p_va);

    if (!p_va->conn)
        abort();

    p_va->conn->lock();

    bool ready = false;
    VAStatus status;
    VASurfaceStatus surface;
    VASurfaceID id = pic->p_sys->i_id;
    status = vaQuerySurfaceStatus(p_va->conn->p_display, id, &surface);
    if ((status == VA_STATUS_SUCCESS) &&
        (surface == VASurfaceReady))
        ready = true;

    p_va->conn->unlock();

    return ready;
}

/* */
int vlc_va_New( vlc_va_t *p_va, int pixfmt, int i_codec_id,
                const es_format_t *fmt )
{
    /* Only VLD supported */
    if( pixfmt != PIX_FMT_VAAPI_VLD )
        return VLC_EGENERIC;

    if( !vlc_xlib_init( VLC_OBJECT(p_va) ) )
    {
        msg_Warn( p_va, "Ignoring VA API" );
        return VLC_EGENERIC;
    }

    (void) fmt;

    int err = Open( p_va, i_codec_id );
    if( err )
        return err;

    /* */
    p_va->direct_rendering = p_va->conn->b_direct_rendering;
    p_va->setup = Setup;
    p_va->get = Get;
    p_va->put = Put;
    p_va->release = Release;
    p_va->extract = Extract;
    p_va->display = DisplayPicture;
    p_va->query = QuerySurfaceReady;
    p_va->close = Delete;
    return VLC_SUCCESS;
}

